@@ -32,15 +32,17 @@ package main
 
 import (
 	"bufio"
+	"cmp"
 	"eutils"
 	"fmt"
 	"io"
+	"maps"
 	"math"
 	"os"
 	"runtime"
 	"runtime/debug"
 	"runtime/pprof"
-	"sort"
+	"slices"
 	"strconv"
 	"strings"
 	"time"
@@ -79,7 +81,7 @@ func parseMarkup(str, cmd string) int {
 		return eutils.TERSE
 	default:
 		if str != "" {
-			fmt.Fprintf(os.Stderr, "\nERROR: Unrecognized %s value '%s'\n", cmd, str)
+			eutils.DisplayError("Unrecognized %s value '%s'", cmd, str)
 			os.Exit(1)
 		}
 	}
@@ -97,10 +99,16 @@ func main() {
 	args := os.Args[1:]
 
 	if len(args) < 1 {
-		fmt.Fprintf(os.Stderr, "\nERROR: No command-line arguments supplied to xtract\n")
+		eutils.DisplayError("No command-line arguments supplied to xtract")
 		os.Exit(1)
 	}
 
+	// pretty-printing of xtract arguments
+	if args[0] == "-pretty" {
+		eutils.PrettyArguments(args[1:])
+		return
+	}
+
 	// performance arguments
 	chanDepth := 0
 	farmSize := 0
@@ -302,7 +310,7 @@ func main() {
 	case "none", "default":
 	default:
 		if flgs != "" {
-			fmt.Fprintf(os.Stderr, "\nERROR: Unrecognized -flag value '%s'\n", flgs)
+			eutils.DisplayError("Unrecognized -flag value '%s'", flgs)
 			os.Exit(1)
 		}
 	}
@@ -352,7 +360,7 @@ func main() {
 	}
 
 	if len(args) < 1 {
-		fmt.Fprintf(os.Stderr, "\nERROR: Insufficient command-line arguments supplied to xtract\n")
+		eutils.DisplayError("Insufficient command-line arguments supplied to xtract")
 		os.Exit(1)
 	}
 
@@ -401,7 +409,7 @@ func main() {
 
 		inFile, err := os.Open(fileName)
 		if err != nil {
-			fmt.Fprintf(os.Stderr, "\nERROR: Unable to open input file '%s'\n", fileName)
+			eutils.DisplayError("Unable to open input file '%s'", fileName)
 			os.Exit(1)
 		}
 
@@ -413,7 +421,7 @@ func main() {
 
 		if isPipe && runtime.GOOS != "windows" {
 			mode := fi.Mode().String()
-			fmt.Fprintf(os.Stderr, "\nERROR: Input data from both stdin and file '%s', mode is '%s'\n", fileName, mode)
+			eutils.DisplayError("Input data from both stdin and file '%s', mode is '%s'", fileName, mode)
 			os.Exit(1)
 		}
 	}
@@ -421,7 +429,7 @@ func main() {
 	// check for -input command after extraction arguments
 	for _, str := range args {
 		if str == "-input" {
-			fmt.Fprintf(os.Stderr, "\nERROR: Misplaced -input command\n")
+			eutils.DisplayError("Misplaced -input command")
 			os.Exit(1)
 		}
 	}
@@ -432,7 +440,7 @@ func main() {
 
 		f, err := os.Create("cpu.pprof")
 		if err != nil {
-			fmt.Fprintf(os.Stderr, "\nERROR: Unable to create profile output file\n")
+			eutils.DisplayError("Unable to create profile output file")
 			os.Exit(1)
 		}
 
@@ -461,7 +469,7 @@ func main() {
 
 		inFile, err := os.Open(tf)
 		if err != nil {
-			fmt.Fprintf(os.Stderr, "Unable to open transformation file %s\n", err.Error())
+			eutils.DisplayError("Unable to open transformation file %s", err.Error())
 			os.Exit(1)
 		}
 		defer inFile.Close()
@@ -508,23 +516,29 @@ func main() {
 	// SEQUENCE RECORD EXTRACTION COMMAND GENERATOR
 
 	// -insd simplifies extraction of INSDSeq qualifiers
-	if args[0] == "-insd" || args[0] == "-insd-" || args[0] == "-insd-idx" {
+	if args[0] == "-insd" || args[0] == "-insd-" || args[0] == "-insd-idx" || args[0] == "-insd-xml" {
 
 		addDash := true
 		doIndex := false
+		makeXML := false
+
 		// -insd- variant suppresses use of dash as placeholder for missing qualifiers (undocumented)
 		if args[0] == "-insd-" {
 			addDash = false
 		}
-		// -insd-idx variant creates word index using -indices command (undocumented)
+		// -insd-idx variant creates word index using the -wrp TIAB -indexer command (undocumented)
 		if args[0] == "-insd-idx" {
 			doIndex = true
 			addDash = false
 		}
+		// -insd-xml variant used for -insdx and subsequent conversion from table to XML (undocumented)
+		if args[0] == "-insd-xml" {
+			makeXML = true
+		}
 
 		args = args[1:]
 
-		insd := eutils.ProcessINSD(args, isPipe || usingFile, addDash, doIndex)
+		insd := eutils.ProcessINSD(args, isPipe || usingFile, addDash, doIndex, makeXML)
 
 		if !isPipe && !usingFile {
 			// no piped input, so write output instructions
@@ -540,37 +554,6 @@ func main() {
 		args = insd
 	}
 
-	// CITATION MATCHER EXTRACTION COMMAND GENERATOR
-
-	// -citmatch extracts PMIDs from nquire -citmatch output (undocumented)
-	if args[0] == "-citmatch" {
-
-		var acc []string
-
-		acc = append(acc, "-pattern", "opt")
-		acc = append(acc, "-if", "success", "-equals", "true")
-		acc = append(acc, "-and", "result/count", "-eq", "1")
-		if isPipe {
-			acc = append(acc, "-sep", "\n")
-		} else {
-			acc = append(acc, "-sep", "\"\\n\"")
-		}
-		acc = append(acc, "-element", "uids/pubmed")
-
-		if !isPipe && !usingFile {
-			// no piped input, so write output instructions
-			fmt.Printf("xtract")
-			for _, str := range acc {
-				fmt.Printf(" %s", str)
-			}
-			fmt.Printf("\n")
-			return
-		}
-
-		// data in pipe, so replace arguments, execute dynamically
-		args = acc
-	}
-
 	// BIOTHINGS EXTRACTION COMMAND GENERATOR
 
 	// -biopath takes a parent object and a dotted exploration path for BioThings resources (undocumented)
@@ -609,7 +592,7 @@ func main() {
 		switch args[0] {
 		case "-head":
 			if len(args) < 2 {
-				fmt.Fprintf(os.Stderr, "\nERROR: Pattern missing after -head command\n")
+				eutils.DisplayError("Pattern missing after -head command")
 				os.Exit(1)
 			}
 			head = eutils.ConvertSlash(args[1])
@@ -636,26 +619,26 @@ func main() {
 			}
 		case "-tail":
 			if len(args) < 2 {
-				fmt.Fprintf(os.Stderr, "\nERROR: Pattern missing after -tail command\n")
+				eutils.DisplayError("Pattern missing after -tail command")
 				os.Exit(1)
 			}
 			tail = eutils.ConvertSlash(args[1])
 		case "-hd":
 			if len(args) < 2 {
-				fmt.Fprintf(os.Stderr, "\nERROR: Pattern missing after -hd command\n")
+				eutils.DisplayError("Pattern missing after -hd command")
 				os.Exit(1)
 			}
 			hd = eutils.ConvertSlash(args[1])
 		case "-tl":
 			if len(args) < 2 {
-				fmt.Fprintf(os.Stderr, "\nERROR: Pattern missing after -tl command\n")
+				eutils.DisplayError("Pattern missing after -tl command")
 				os.Exit(1)
 			}
 			tl = eutils.ConvertSlash(args[1])
 		case "-wrp":
 			// shortcut to wrap records in XML tags
 			if len(args) < 2 {
-				fmt.Fprintf(os.Stderr, "\nERROR: Pattern missing after -wrp command\n")
+				eutils.DisplayError("Pattern missing after -wrp command")
 				os.Exit(1)
 			}
 			tmp := eutils.ConvertSlash(args[1])
@@ -670,7 +653,7 @@ func main() {
 			}
 		case "-set":
 			if len(args) < 2 {
-				fmt.Fprintf(os.Stderr, "\nERROR: Pattern missing after -set command\n")
+				eutils.DisplayError("Pattern missing after -set command")
 				os.Exit(1)
 			}
 			tmp := eutils.ConvertSlash(args[1])
@@ -680,7 +663,7 @@ func main() {
 			}
 		case "-rec":
 			if len(args) < 2 {
-				fmt.Fprintf(os.Stderr, "\nERROR: Pattern missing after -rec command\n")
+				eutils.DisplayError("Pattern missing after -rec command")
 				os.Exit(1)
 			}
 			tmp := eutils.ConvertSlash(args[1])
@@ -701,7 +684,7 @@ func main() {
 		args = args[2:]
 
 		if len(args) < 1 {
-			fmt.Fprintf(os.Stderr, "\nERROR: Insufficient command-line arguments supplied to xtract\n")
+			eutils.DisplayError("Insufficient command-line arguments supplied to xtract")
 			os.Exit(1)
 		}
 	}
@@ -715,7 +698,7 @@ func main() {
 		buffer := make([]byte, FirstBuffSize)
 		n, err := in.Read(buffer)
 		if err != nil && err != io.EOF {
-			fmt.Fprintf(os.Stderr, "\nERROR: Unable to read first block: %s\n", err.Error())
+			eutils.DisplayError("Unable to read first block: %s", err.Error())
 			// os.Exit(1)
 		}
 		bufr := buffer[:n]
@@ -724,6 +707,9 @@ func main() {
 
 	first := getFirstBlock()
 
+	// trim leading byte order mark, if present
+	first = eutils.TrimBOM(first)
+
 	mlt := io.MultiReader(strings.NewReader(first), in)
 
 	isJsn := false
@@ -735,9 +721,13 @@ func main() {
 	if first != "" {
 		posJ1 := strings.Index(first, "{")
 		posJ2 := strings.Index(first, "\":")
+		posJ3 := strings.Index(first, "\" :")
 		if posJ1 >= 0 && posJ2 >= 0 && posJ1 < posJ2 {
 			isJsn = true
 			matched++
+		} else if posJ1 >= 0 && posJ3 >= 0 && posJ1 < posJ3 {
+			isJsn = true
+			matched++
 		} else {
 			posJ1 = FirstBuffSize
 			posJ2 = FirstBuffSize
@@ -786,20 +776,22 @@ func main() {
 		}
 	}
 
+	var xmlChan <-chan string
+
 	if isJsn {
-		jrdr := eutils.JSONConverter(mlt, "root", "", "element")
-		mlt = eutils.ChanToReader(jrdr)
+		xmlChan = eutils.JSONConverter(mlt, "root", "", "element")
+		mlt = nil
 	} else if isAsn {
-		ardr := eutils.ASN1Converter(mlt, "", "")
-		mlt = eutils.ChanToReader(ardr)
+		xmlChan = eutils.ASN1Converter(mlt, "", "")
+		mlt = nil
 	} else if isGbf {
-		grdr := eutils.GenBankConverter(mlt)
-		mlt = eutils.ChanToReader(grdr)
+		xmlChan = eutils.GenBankConverter(mlt)
+		mlt = nil
 	}
 
-	rdr := eutils.CreateXMLStreamer(mlt)
+	rdr := eutils.CreateXMLStreamer(mlt, xmlChan)
 	if rdr == nil {
-		fmt.Fprintf(os.Stderr, "\nERROR: Unable to create XML Block Reader\n")
+		eutils.DisplayError("Unable to create XML Block Reader")
 		os.Exit(1)
 	}
 
@@ -810,14 +802,14 @@ func main() {
 		fromStdin := bool((fi.Mode() & os.ModeCharDevice) == 0)
 		if !isPipe || !fromStdin {
 			mode := fi.Mode().String()
-			fmt.Fprintf(os.Stderr, "\nERROR: No data supplied to xtract from stdin or file, mode is '%s'\n", mode)
+			eutils.DisplayError("No data supplied to xtract from stdin or file, mode is '%s'", mode)
 			os.Exit(1)
 		}
 	}
 
 	if !usingFile && !isPipe {
 
-		fmt.Fprintf(os.Stderr, "\nERROR: No XML input data supplied to xtract\n")
+		eutils.DisplayError("No XML input data supplied to xtract")
 		os.Exit(1)
 	}
 
@@ -843,6 +835,7 @@ func main() {
 
 		find := ""
 		html := false
+		max := 0
 
 		// look for optional arguments
 		for {
@@ -853,14 +846,21 @@ func main() {
 
 			switch arg {
 			case "-find":
-				// override set wrapper
-				find, ok = nextArg()
+				// set identifier object
+				find, _ = nextArg()
 			case "-html":
 				html = true
+			case "-max":
+				// override reportable depth
+				cutoff, _ := nextArg()
+				val, err := strconv.Atoi(cutoff)
+				if err == nil && val > 0 {
+					max = val
+				}
 			}
 		}
 
-		recordCount = eutils.ValidateXML(rdr, find, html)
+		recordCount = eutils.ValidateXML(rdr, find, html, max)
 
 		debug.FreeOSMemory()
 
@@ -899,7 +899,7 @@ func main() {
 			}
 		}
 		if len(args) < 2 {
-			fmt.Fprintf(os.Stderr, "\nERROR: Pattern missing after -split command\n")
+			eutils.DisplayError("Pattern missing after -split command")
 			os.Exit(1)
 		}
 		pat := args[1]
@@ -941,7 +941,7 @@ func main() {
 			}
 		}
 		if len(args) < 2 {
-			fmt.Fprintf(os.Stderr, "\nERROR: Pattern missing after -index command\n")
+			eutils.DisplayError("Pattern missing after -index command")
 			os.Exit(1)
 		}
 		pat := args[1]
@@ -993,7 +993,7 @@ func main() {
 	// ENSURE PRESENCE OF PATTERN ARGUMENT
 
 	if len(args) < 1 {
-		fmt.Fprintf(os.Stderr, "\nERROR: Insufficient command-line arguments supplied to xtract\n")
+		eutils.DisplayError("Insufficient command-line arguments supplied to xtract")
 		os.Exit(1)
 	}
 
@@ -1004,21 +1004,21 @@ func main() {
 
 	// make sure top-level -pattern command is next
 	if args[0] != "-pattern" && args[0] != "-Pattern" {
-		fmt.Fprintf(os.Stderr, "\nERROR: No -pattern in command-line arguments\n")
+		eutils.DisplayError("No -pattern in command-line arguments")
 		os.Exit(1)
 	}
 	if len(args) < 2 {
-		fmt.Fprintf(os.Stderr, "\nERROR: Item missing after -pattern command\n")
+		eutils.DisplayError("Item missing after -pattern command")
 		os.Exit(1)
 	}
 
 	topPat := args[1]
 	if topPat == "" {
-		fmt.Fprintf(os.Stderr, "\nERROR: Item missing after -pattern command\n")
+		eutils.DisplayError("Item missing after -pattern command")
 		os.Exit(1)
 	}
 	if strings.HasPrefix(topPat, "-") {
-		fmt.Fprintf(os.Stderr, "\nERROR: Misplaced %s command\n", topPat)
+		eutils.DisplayError("Misplaced %s command", topPat)
 		os.Exit(1)
 	}
 
@@ -1032,7 +1032,7 @@ func main() {
 	if star == "*" {
 		parent = topPattern
 	} else if star != "" {
-		fmt.Fprintf(os.Stderr, "\nERROR: -pattern Parent/Child construct is not supported\n")
+		eutils.DisplayError("-pattern Parent/Child construct is not supported")
 		os.Exit(1)
 	}
 
@@ -1046,7 +1046,7 @@ func main() {
 		unsq := eutils.CreateXMLUnshuffler(fchq)
 
 		if xmlq == nil || fchq == nil || unsq == nil {
-			fmt.Fprintf(os.Stderr, "\nERROR: Unable to create selector\n")
+			eutils.DisplayError("Unable to create selector")
 			os.Exit(1)
 		}
 
@@ -1109,7 +1109,7 @@ func main() {
 		// read file of identifiers to use for filtering
 		fl, err := os.Open(unqe)
 		if err != nil {
-			fmt.Fprintf(os.Stderr, "\nERROR: Unable to open identifier file '%s'\n", unqe)
+			eutils.DisplayError("Unable to open identifier file '%s'", unqe)
 			os.Exit(1)
 		}
 
@@ -1137,7 +1137,7 @@ func main() {
 		unsq := eutils.CreateXMLUnshuffler(fchq)
 
 		if xmlq == nil || fchq == nil || unsq == nil {
-			fmt.Fprintf(os.Stderr, "\nERROR: Unable to create selector\n")
+			eutils.DisplayError("Unable to create selector")
 			os.Exit(1)
 		}
 
@@ -1200,7 +1200,7 @@ func main() {
 		// read file of identifiers to use for filtering
 		fl, err := os.Open(unqe)
 		if err != nil {
-			fmt.Fprintf(os.Stderr, "\nERROR: Unable to open identifier file '%s'\n", unqe)
+			eutils.DisplayError("Unable to open identifier file '%s'", unqe)
 			os.Exit(1)
 		}
 
@@ -1282,7 +1282,7 @@ func main() {
 
 		fl, err := os.Open(apnd)
 		if err != nil {
-			fmt.Fprintf(os.Stderr, "\nERROR: Unable to open transformation file '%s'\n", apnd)
+			eutils.DisplayError("Unable to open transformation file '%s'", apnd)
 			os.Exit(1)
 		}
 
@@ -1391,57 +1391,41 @@ func main() {
 				order[id] = data
 			})
 
-		var keys []string
-		for ky := range order {
-			keys = append(keys, ky)
-		}
+		// sort fields in alphabetical or numeric order, modified from sort.Slice (Less) to slices.SortedFunc (Compare) style
+		xtractSort := func(i, j string) int {
 
-		// sort fields in alphabetical or numeric order
-		sortKeys := func(i, j int) bool {
-			// numeric sort on strings checks lengths first
-			if eutils.IsAllDigits(keys[i]) && eutils.IsAllDigits(keys[j]) {
-				lni := len(keys[i])
-				lnj := len(keys[j])
-				// shorter string is numerically less, assuming no leading zeros
+			// check for numeric (integer or decimal) representation in both string arguments
+			if eutils.IsAllDigitsOrPeriod(i) && eutils.IsAllDigitsOrPeriod(j) {
+
+				// isolate integer and fractional component if decimal number
+				ni, _ := eutils.SplitInTwoLeft(i, ".")
+				nj, _ := eutils.SplitInTwoLeft(j, ".")
+				// get lengths of whole number
+				lni, lnj := len(ni), len(nj)
+				// shorter string is numerically lower, assuming no leading zeros
 				if lni < lnj {
-					return true
+					return -1
 				}
 				if lni > lnj {
-					return false
+					return 1
 				}
+				// integer components are same length, thus right (or decimal point) aligned for text
+				// comparison of numeric values without needing expensive string-to-integer conversion
 			}
-			// real numbers, split at decimal point
-			if eutils.IsAllDigitsOrPeriod(keys[i]) && eutils.IsAllDigitsOrPeriod(keys[j]) {
-				lfti, rgti := eutils.SplitInTwoLeft(keys[i], ".")
-				lftj, rgtj := eutils.SplitInTwoLeft(keys[j], ".")
-				lni := len(lfti)
-				lnj := len(lftj)
-				// shorter string is numerically less, assuming no leading zeros
-				if lni < lnj {
-					return true
-				}
-				if lni > lnj {
-					return false
-				}
-				if lfti != lftj {
-					// compare integer portion
-					return lfti < lftj
-				}
-				// compare decimal portion
-				return rgti < rgtj
-			}
-			// same length or non-numeric, can now do string comparison on contents
-			return keys[i] < keys[j]
+
+			// same length integers or non-numeric, can now do string comparison on contents
+			return cmp.Compare(i, j)
 		}
 
-		// return sorted results in in forward or reverse order
-		sort.Slice(keys, func(i, j int) bool {
-			res := sortKeys(i, j)
-			if sortInReverse {
-				res = !res
+		sortKeys := xtractSort
+		if sortInReverse {
+			sortKeys = func(i, j string) int {
+				// sends arguments in reverse order
+				return xtractSort(j, i)
 			}
-			return res
-		})
+		}
+
+		keys := slices.SortedFunc(maps.Keys(order), sortKeys)
 
 		if head != "" {
 			os.Stdout.WriteString(head)
@@ -1485,7 +1469,7 @@ func main() {
 		)
 		chunk, err := strconv.Atoi(args[3])
 		if err != nil {
-			fmt.Fprintf(os.Stderr, "%s\n", err.Error())
+			eutils.DisplayError("-split argument '%s' is not an integer", err.Error())
 			return
 		}
 		prefix := args[5]
@@ -1507,7 +1491,7 @@ func main() {
 					fpath := fmt.Sprintf("%s%03d.%s", prefix, fnum, suffix)
 					fl, err = os.Create(fpath)
 					if err != nil {
-						fmt.Fprintf(os.Stderr, "%s\n", err.Error())
+						eutils.DisplayError("Unable to create path '%s'", err.Error())
 						return
 					}
 					os.Stderr.WriteString(fpath + "\n")
@@ -1523,7 +1507,7 @@ func main() {
 				fl.WriteString("\n")
 			})
 
-		if count >= chunk {
+		if count > 0 {
 			if tail != "" {
 				fl.WriteString(tail)
 				fl.WriteString("\n")
@@ -1545,7 +1529,7 @@ func main() {
 	// parse nested exploration instruction from command-line arguments
 	cmds := eutils.ParseArguments(args, topPattern)
 	if cmds == nil {
-		fmt.Fprintf(os.Stderr, "\nERROR: Problem parsing command-line arguments\n")
+		eutils.DisplayError("Problem parsing command-line arguments")
 		os.Exit(1)
 	}
 
@@ -1562,9 +1546,13 @@ func main() {
 
 		rec := 0
 
+		// find := eutils.ParseIndex("MedlineCitation/PMID")
+
 		eutils.PartitionXML(topPattern, star, turbo, rdr,
 			func(str string) {
 				rec++
+				// id := eutils.FindIdentifier(str[:], parent, find)
+				// if eutils.HasCombiningAccent(str[:]) && id != "" { fmt.Printf("%s\n", id) }
 				beginTime := time.Now()
 				eutils.ProcessExtract(str[:], parent, rec, hd, tl, transform, nil, histogram, cmds)
 				endTime := time.Now()
@@ -1601,17 +1589,17 @@ func main() {
 			m2 := 0.0
 
 			// calculate mean and standard deviation of processing rate
-			for trials := 0; trials < 5; trials++ {
+			for range 5 {
 
 				inFile, err := os.Open(fileName)
 				if err != nil {
-					fmt.Fprintf(os.Stderr, "\nERROR: Unable to open input file '%s'\n", fileName)
+					eutils.DisplayError("Unable to open input file '%s'", fileName)
 					os.Exit(1)
 				}
 
-				trdr := eutils.CreateXMLStreamer(inFile)
+				trdr := eutils.CreateXMLStreamer(inFile, nil)
 				if trdr == nil {
-					fmt.Fprintf(os.Stderr, "\nERROR: Unable to read input file\n")
+					eutils.DisplayError("Unable to read input file")
 					os.Exit(1)
 				}
 
@@ -1619,7 +1607,7 @@ func main() {
 				tblq := eutils.CreateXMLConsumers(cmds, parent, hd, tl, transform, false, histogram, xmlq)
 
 				if xmlq == nil || tblq == nil {
-					fmt.Fprintf(os.Stderr, "\nERROR: Unable to create servers\n")
+					eutils.DisplayError("Unable to create servers")
 					os.Exit(1)
 				}
 
@@ -1709,7 +1697,7 @@ func main() {
 			// use numeric position
 			number, err := strconv.Atoi(cmds.Position)
 			if err != nil {
-				fmt.Fprintf(os.Stderr, "\nERROR: Unrecognized position '%s'\n", cmds.Position)
+				eutils.DisplayError("Unrecognized position '%s'", cmds.Position)
 				os.Exit(1)
 			}
 
@@ -1752,7 +1740,7 @@ func main() {
 	unsq := eutils.CreateXMLUnshuffler(tblq)
 
 	if xmlq == nil || tblq == nil || unsq == nil {
-		fmt.Fprintf(os.Stderr, "\nERROR: Unable to create servers\n")
+		eutils.DisplayError("Unable to create servers")
 		os.Exit(1)
 	}
 
